from .exceptions import UnexpectedMethodCallError


class MethodGroup(object):
    """Base class containing common behaviour for MethodGroups."""

    def __init__(self, group_name, exception_list):
        """Construct a new method group.

        Args:
          # group_name: the name of the method group
          # exception_list: list of exceptions; any exceptions thrown by this
          #     instance are appended to this list.
          group_name: str
          exception_list: list
        """
        self._group_name = group_name
        self._exception_list = exception_list

    def group_name(self):
        return self._group_name

    def __str__(self):
        return '<%s "%s">' % (self.__class__.__name__, self._group_name)

    def add_method(self, mock_method):
        raise NotImplementedError

    def method_called(self, mock_method):
        raise NotImplementedError

    def is_satisfied(self):
        raise NotImplementedError

    AddMethod = add_method
    MethodCalled = method_called
    IsSatisfied = is_satisfied


class UnorderedGroup(MethodGroup):
    """UnorderedGroup holds a set of method calls that may occur in any order.

    This construct is helpful for non-deterministic events, such as iterating
    over the keys of a dict.
    """

    def __init__(self, group_name, exception_list):
        super(UnorderedGroup, self).__init__(group_name, exception_list)
        self._methods = []

    def __str__(self):
        return '%s "%s" pending calls:\n%s' % (
            self.__class__.__name__,
            self._group_name,
            "\n".join(str(method) for method in self._methods),
        )

    def add_method(self, mock_method):
        """Add a method to this group.

        Args:
          mock_method: A mock method to be added to this group.
        """

        self._methods.append(mock_method)

    def method_called(self, mock_method):
        """Remove a method call from the group.

        If the method is not in the set, an UnexpectedMethodCallError will be
        raised.

        Args:
          mock_method: a mock method that should be equal to a method in the
          group.

        Returns:
          The mock method from the group

        Raises:
          UnexpectedMethodCallError if the mock_method was not in the group.
        """

        # Check to see if this method exists, and if so, remove it from the set
        # and return it.
        for method in self._methods:
            if method == mock_method:
                # Remove the called mock_method instead of the method in the
                # group. The called method will match any comparators when
                # equality is checked during removal. The method in the group
                # could pass a comparator to another comparator during the
                # equality check.
                self._methods.remove(mock_method)

                # If this group is not empty, put it back at the head of the
                # queue.
                if not self.is_satisfied():
                    mock_method._call_queue.appendleft(self)

                return self, method

        exception = UnexpectedMethodCallError(mock_method, self)
        self._exception_list.append(exception)
        raise exception

    def is_satisfied(self):
        """Return True if there are not any methods in this group."""

        return len(self._methods) == 0

    AddMethod = add_method
    MethodCalled = method_called
    IsSatisfied = is_satisfied


class MultipleTimesGroup(MethodGroup):
    """MultipleTimesGroup holds methods that may be called any number of times.

    Note: Each method must be called at least once.

    This is helpful, if you don't know or care how many times a method is
    called.
    """

    def __init__(self, group_name, exception_list):
        super(MultipleTimesGroup, self).__init__(group_name, exception_list)
        self._methods = set()
        self._methods_left = set()

    def add_method(self, mock_method):
        """Add a method to this group.

        Args:
          mock_method: A mock method to be added to this group.
        """

        self._methods.add(mock_method)
        self._methods_left.add(mock_method)

    def method_called(self, mock_method):
        """Remove a method call from the group.

        If the method is not in the set, an UnexpectedMethodCallError will be
        raised.

        Args:
          mock_method: a mock method that should be equal to a method in the
            group.

        Returns:
          The mock method from the group

        Raises:
          UnexpectedMethodCallError if the mock_method was not in the group.
        """

        # Check to see if this method exists, and if so add it to the set of
        # called methods.
        for method in self._methods:
            if method == mock_method:
                self._methods_left.discard(method)
                # Always put this group back on top of the queue, because we
                # don't know when we are done.
                mock_method._call_queue.appendleft(self)
                return self, method

        if self.is_satisfied():
            next_method = mock_method._pop_next_method()
            return next_method, None
        else:
            exception = UnexpectedMethodCallError(mock_method, self)
            self._exception_list.append(exception)
            raise exception

    def is_satisfied(self):
        """Return True if all methods in this group are called at least once."""
        return len(self._methods_left) == 0

    AddMethod = add_method
    MethodCalled = method_called
    IsSatisfied = is_satisfied



#!/usr/bin/env python
#
# Copyright 2008 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Python imports
import inspect


class StubOutForTesting:
    """Sample Usage:
    You want os.path.exists() to always return true during testing.

    stubs = StubOutForTesting()
    stubs.set(os.path, 'exists', lambda x: 1)
      ...
    stubs.unset_all()

    The above changes os.path.exists into a lambda that returns 1. Once
    the ... part of the code finishes, the unset_all() looks up the old value
    of os.path.exists and restores it.

    """

    def __init__(self):
        self.cache = []
        self.stubs = []

    def __del__(self):
        self.smart_unset_all()
        self.unset_all()

    def smart_set(self, obj, attr_name, new_attr):
        """Replace obj.attr_name with new_attr. This method is smart and works
         at the module, class, and instance level while preserving proper
         inheritance. It will not stub out C types however unless that has
         been explicitly allowed by the type.

         This method supports the case where attr_name is a staticmethod or a
         classmethod of obj.

         Notes:
        - If obj is an instance, then it is its class that will actually be
          stubbed. Note that the method Set() does not do that: if obj is
          an instance, it (and not its class) will be stubbed.
        - The stubbing is using the builtin getattr and setattr. So, the
          __get__ and __set__ will be called when stubbing (TODO: A better
          idea would probably be to manipulate obj.__dict__ instead of
          getattr() and setattr()).

         Raises AttributeError if the attribute cannot be found.
        """
        if inspect.ismodule(obj) or (not inspect.isclass(obj) and attr_name in obj.__dict__):
            orig_obj = obj
            orig_attr = getattr(obj, attr_name)

        else:
            if not inspect.isclass(obj):
                mro = list(inspect.getmro(obj.__class__))
            else:
                mro = list(inspect.getmro(obj))

            mro.reverse()

            orig_attr = None

            for cls in mro:
                try:
                    orig_obj = cls
                    orig_attr = getattr(obj, attr_name)
                except AttributeError:
                    continue

        if orig_attr is None:
            raise AttributeError("Attribute not found.")

        # Calling getattr() on a staticmethod transforms it to a 'normal'
        # function. We need to ensure that we put it back as a staticmethod.
        old_attribute = obj.__dict__.get(attr_name)
        if old_attribute is not None and isinstance(old_attribute, staticmethod):
            orig_attr = staticmethod(orig_attr)

        self.stubs.append((orig_obj, attr_name, orig_attr))
        setattr(orig_obj, attr_name, new_attr)

    def smart_unset_all(self):
        """Reverses all the SmartSet() calls, restoring things to their
        original definition. It's okay to call smart_unset_all() repeatedly, as
        later calls have no effect if no SmartSet() calls have been made.

        """
        self.stubs.reverse()

        for args in self.stubs:
            setattr(*args)

        self.stubs = []

    def set(self, parent, child_name, new_child):
        """Replace child_name's old definition with new_child, in the context
        of the given parent.  The parent could be a module when the child is a
        function at module scope.  Or the parent could be a class when a class'
        method is being replaced.  The named child is set to new_child, while
        the prior definition is saved away for later, when unset_all() is
        called.

        This method supports the case where child_name is a staticmethod or a
        classmethod of parent.
        """
        old_child = getattr(parent, child_name)

        old_attribute = parent.__dict__.get(child_name)
        if old_attribute is not None:
            if isinstance(old_attribute, staticmethod):
                old_child = staticmethod(old_child)
            elif isinstance(old_attribute, classmethod):
                old_child = classmethod(old_child.__func__)

        self.cache.append((parent, old_child, child_name))
        setattr(parent, child_name, new_child)

    def unset_all(self):
        """Reverses all the set() calls, restoring things to their original
        definition. It's okay to call unset_all() repeatedly, as later calls
        have no effect if no set() calls have been made.

        """
        # Undo calls to set() in reverse order, in case set() was called on the
        # same arguments repeatedly (want the original call to be last one
        # undone)
        self.cache.reverse()

        for parent, old_child, child_name in self.cache:
            setattr(parent, child_name, old_child)
        self.cache = []

    SmartSet = smart_set
    SmartUnsetAll = smart_unset_all
    Set = set
    UnsetAll = unset_all


stubout = StubOutForTesting



# Python imports
import re

from .exceptions import Error


class Comparator:
    """Base class for all Mox comparators.

    A Comparator can be used as a parameter to a mocked method when the exact
    value is not known.  For example, the code you are testing might build up a
    long SQL string that is passed to your mock DAO. You're only interested
    that the IN clause contains the proper primary keys, so you can set your
    mock up as follows:

    mock_dao.RunQuery(StrContains('IN (1, 2, 4, 5)')).AndReturn(mock_result)

    Now whatever query is passed in must contain the string 'IN (1, 2, 4, 5)'.

    A Comparator may replace one or more parameters, for example:
    # return at most 10 rows
    mock_dao.RunQuery(StrContains('SELECT'), 10)

    or

    # Return some non-deterministic number of rows
    mock_dao.RunQuery(StrContains('SELECT'), IsA(int))
    """

    def equals(self, rhs):
        """Special equals method that all comparators must implement.

        Args:
          rhs: any python object
        """

        raise NotImplementedError

    def __eq__(self, rhs):
        return self.equals(rhs)

    def __ne__(self, rhs):
        return not self.equals(rhs)


class Is(Comparator):
    """Comparison class used to check identity, instead of equality."""

    def __init__(self, obj):
        self._obj = obj

    def equals(self, rhs):
        return rhs is self._obj

    def __repr__(self):
        return "<is %r (%s)>" % (self._obj, id(self._obj))


class IsA(Comparator):
    """This class wraps a basic Python type or class.  It is used to verify
    that a parameter is of the given type or class.

    Example:
    mock_dao.Connect(IsA(DbConnectInfo))
    """

    def __init__(self, class_name):
        """Initialize IsA

        Args:
          class_name: basic python type or a class
        """

        self._class_name = class_name

    def equals(self, rhs):
        """Check to see if the RHS is an instance of class_name.

        Args:
          # rhs: the right hand side of the test
          rhs: object

        Returns:
          bool
        """

        try:
            return isinstance(rhs, self._class_name)
        except TypeError:
            # Check raw types if there was a type error.  This is helpful for
            # things like cStringIO.StringIO.
            return isinstance(rhs, type(self._class_name))

    def _is_subclass(self, clazz):
        """Check to see if the IsA comparators class is a subclass of clazz.

        Args:
          # clazz: a class object

        Returns:
          bool
        """

        try:
            return issubclass(self._class_name, clazz)
        except TypeError:
            # Check raw types if there was a type error.  This is helpful for
            # things like cStringIO.StringIO.
            return isinstance(clazz, type(self._class_name))

    def __repr__(self):
        return "mox.IsA(%s) " % str(self._class_name)

    _IsSubClass = _is_subclass


class IsAlmost(Comparator):
    """Comparison class used to check whether a parameter is nearly equal
    to a given value.  Generally useful for floating point numbers.

    Example mock_dao.SetTimeout((IsAlmost(3.9)))
    """

    def __init__(self, float_value, places=7):
        """Initialize IsAlmost.

        Args:
          float_value: The value for making the comparison.
          places: The number of decimal places to round to.
        """

        self._float_value = float_value
        self._places = places

    def equals(self, rhs):
        """Check to see if RHS is almost equal to float_value

        Args:
          rhs: the value to compare to float_value

        Returns:
          bool
        """

        try:
            return round(rhs - self._float_value, self._places) == 0
        except Exception:
            # This is probably because either float_value or rhs is not a
            # number.
            return False

    def __repr__(self):
        return str(self._float_value)


class StrContains(Comparator):
    """Comparison class used to check whether a substring exists in a
    string parameter.  This can be useful in mocking a database with SQL
    passed in as a string parameter, for example.

    Example:
    mock_dao.RunQuery(StrContains('IN (1, 2, 4, 5)')).AndReturn(mock_result)
    """

    def __init__(self, search_string):
        """Initialize.

        Args:
          # search_string: the string you are searching for
          search_string: str
        """

        self._search_string = search_string

    def equals(self, rhs):
        """Check to see if the search_string is contained in the rhs string.

        Args:
          # rhs: the right hand side of the test
          rhs: object

        Returns:
          bool
        """

        try:
            return rhs.find(self._search_string) > -1
        except Exception:
            return False

    def __repr__(self):
        return "<str containing '%s'>" % self._search_string


class Regex(Comparator):
    """Checks if a string matches a regular expression.

    This uses a given regular expression to determine equality.
    """

    def __init__(self, pattern, flags=0):
        """Initialize.

        Args:
          # pattern is the regular expression to search for
          pattern: str
          # flags passed to re.compile function as the second argument
          flags: int
        """

        self.regex = re.compile(pattern, flags=flags)

    def equals(self, rhs):
        """Check to see if rhs matches regular expression pattern.

        Returns:
          bool
        """

        try:
            return self.regex.search(rhs) is not None
        except Exception:
            return False

    def __repr__(self):
        pattern = self.regex.pattern
        if isinstance(pattern, bytes):
            pattern = pattern.decode()
        s = "<regular expression '{}'".format(pattern)
        if self.regex.flags:
            s += ", flags=%d" % self.regex.flags
        s += ">"
        return s


class In(Comparator):
    """Checks whether an item (or key) is in a list (or dict) parameter.

    Example:
    mock_dao.GetUsersInfo(In('expectedUserName')).AndReturn(mock_result)
    """

    def __init__(self, key):
        """Initialize.

        Args:
          # key is anything that could be in a list or a key in a dict
        """

        self._key = key

    def equals(self, rhs):
        """Check to see whether key is in rhs.

        Args:
          rhs: dict

        Returns:
          bool
        """

        try:
            return self._key in rhs
        except Exception:
            return False

    def __repr__(self):
        return "<sequence or map containing '%s'>" % str(self._key)


class Not(Comparator):
    """Checks whether a predicates is False.

    Example:
      mock_dao.UpdateUsers(Not(ContainsKeyValue('stevepm', stevepm_user_info)))
    """

    def __init__(self, predicate):
        """Initialize.

        Args:
          # predicate: a Comparator instance.
        """

        if not isinstance(predicate, Comparator):
            raise Error("predicate %r must be a Comparator." % predicate)
        self._predicate = predicate

    def equals(self, rhs):
        """Check to see whether the predicate is False.

        Args:
          rhs: A value that will be given in argument of the predicate.

        Returns:
          bool
        """

        try:
            return not self._predicate.equals(rhs)
        except Exception:
            return False

    def __repr__(self):
        return "<not '%s'>" % self._predicate


class ContainsKeyValue(Comparator):
    """Checks whether a key/value pair is in a dict parameter.

    Example:
    mock_dao.UpdateUsers(ContainsKeyValue('stevepm', stevepm_user_info))
    """

    def __init__(self, key, value):
        """Initialize.

        Args:
          # key: a key in a dict
          # value: the corresponding value
        """

        self._key = key
        self._value = value

    def equals(self, rhs):
        """Check whether the given key/value pair is in the rhs dict.

        Returns:
          bool
        """

        try:
            return rhs[self._key] == self._value
        except Exception:
            return False

    def __repr__(self):
        return "<map containing the entry '%s: %s'>" % (
            str(self._key),
            str(self._value),
        )


class ContainsAttributeValue(Comparator):
    """Checks whether a passed parameter contains attributes with a given
    value.

    Example:
    mock_dao.UpdateSomething(ContainsAttribute('stevepm', stevepm_user_info))
    """

    def __init__(self, key, value):
        """Initialize.

        Args:
          # key: an attribute name of an object
          # value: the corresponding value
        """

        self._key = key
        self._value = value

    def equals(self, rhs):
        """Check whether the given attribute has a matching value in the rhs
        object.

        Returns:
          bool
        """

        try:
            return getattr(rhs, self._key) == self._value
        except Exception:
            return False


class SameElementsAs(Comparator):
    """Checks whether sequences contain the same elements (ignoring order).

    Example:
    mock_dao.ProcessUsers(SameElementsAs('stevepm', 'salomaki'))
    """

    def __init__(self, expected_seq):
        """Initialize.

        Args:
          expected_seq: a sequence
        """
        # Store in case expected_seq is an iterator.
        self._expected_list = list(expected_seq)

    def equals(self, actual_seq):
        """Check to see whether actual_seq has same elements as expected_seq.

        Args:
          actual_seq: sequence

        Returns:
          bool
        """
        try:
            # Store in case actual_seq is an iterator.  We potentially iterate
            # twice: once to make the dict, once in the list fallback.
            actual_list = list(actual_seq)
        except TypeError:
            # actual_seq cannot be read as a sequence.
            #
            # This happens because Mox uses __eq__ both to check object
            # equality (in MethodSignatureChecker) and to invoke Comparators.
            return False

        try:
            expected = dict([(element, None) for element in self._expected_list])
            actual = dict([(element, None) for element in actual_list])
        except TypeError:
            # Fall back to slower list-compare if any of the objects are
            # unhashable.
            expected = self._expected_list
            actual = actual_list
            for element in actual:
                if element not in expected:
                    return False
            for element in expected:
                if element not in actual:
                    return False
            return True
        else:
            return set(actual_list) == set(self._expected_list)

    def __repr__(self):
        return "<sequence with same elements as '%s'>" % self._expected_list


class And(Comparator):
    """Evaluates one or more Comparators on RHS and returns an AND of the
    results.
    """

    def __init__(self, *args):
        """Initialize.

        Args:
          *args: One or more Comparator
        """

        self._comparators = args

    def equals(self, rhs):
        """Checks whether all Comparators are equal to rhs.

        Args:
          # rhs: can be anything

        Returns:
          bool
        """

        for comparator in self._comparators:
            if not comparator.equals(rhs):
                return False

        return True

    def __repr__(self):
        return "<AND %s>" % str(self._comparators)


class Or(Comparator):
    """Evaluates one or more Comparators on RHS and returns an OR of the
    results.
    """

    def __init__(self, *args):
        """Initialize.

        Args:
          *args: One or more Mox comparators
        """

        self._comparators = args

    def equals(self, rhs):
        """Checks whether any Comparator is equal to rhs.

        Args:
          # rhs: can be anything

        Returns:
          bool
        """

        for comparator in self._comparators:
            if comparator.equals(rhs):
                return True

        return False

    def __repr__(self):
        return "<OR %s>" % str(self._comparators)


class Func(Comparator):
    """Call a function that should verify the parameter passed in is correct.

    You may need the ability to perform more advanced operations on the
    parameter in order to validate it.  You can use this to have a callable
    validate any parameter. The callable should return either True or False.


    Example:

    def myParamValidator(param):
      # Advanced logic here
      return True

    mock_dao.DoSomething(Func(myParamValidator), true)
    """

    def __init__(self, func):
        """Initialize.

        Args:
          func: callable that takes one parameter and returns a bool
        """

        self._func = func

    def equals(self, rhs):
        """Test whether rhs passes the function test.

        rhs is passed into func.

        Args:
          rhs: any python object

        Returns:
          the result of func(rhs)
        """

        return self._func(rhs)

    def __repr__(self):
        return str(self._func)


class IgnoreArg(Comparator):
    """Ignore an argument.

    This can be used when we don't care about an argument of a method call.

    Example:
    # Check if CastMagic is called with 3 as first arg and 'disappear' as
    third. mymock.CastMagic(3, IgnoreArg(), 'disappear')
    """

    def equals(self, unused_rhs):
        """Ignores arguments and returns True.

        Args:
          unused_rhs: any python object

        Returns:
          always returns True
        """

        return True

    def __repr__(self):
        return "<IgnoreArg>"


class Value(Comparator):
    """Compares argument against a remembered value.

    To be used in conjunction with Remember comparator.  See Remember()
    for example.
    """

    def __init__(self):
        self._value = None
        self._has_value = False

    def store_value(self, rhs):
        self._value = rhs
        self._has_value = True

    def equals(self, rhs):
        if not self._has_value:
            return False
        else:
            return rhs == self._value

    def __repr__(self):
        if self._has_value:
            return "<Value %r>" % self._value
        else:
            return "<Value>"


class Remember(Comparator):
    """Remembers the argument to a value store.

    To be used in conjunction with Value comparator.

    Example:
    # Remember the argument for one method call.
    users_list = Value()
    mock_dao.ProcessUsers(Remember(users_list))

    # Check argument against remembered value.
    mock_dao.ReportUsers(users_list)
    """

    def __init__(self, value_store):
        if not isinstance(value_store, Value):
            raise TypeError("value_store is not an instance of the Value class")
        self._value_store = value_store

    def equals(self, rhs):
        self._value_store.store_value(rhs)
        return True

    def __repr__(self):
        return "<Remember %d>" % id(self._value_store)


is_ = Is
is_a = IsA
is_almost = IsAlmost
str_contains = StrContains
regex = Regex
in_ = In
not_ = Not
contains_key_value = ContainsKeyValue
contains_attribute_value = ContainsAttributeValue
same_elements_as = SameElementsAs
and_ = And
or_ = Or
func = Func
ignore_arg = IgnoreArg
value = Value
remember = Remember
